// PointBillboard.geom

// SetParameter GL_GEOMETRY_INPUT_TYPE_EXT=GL_POINTS
// SetParameter GL_GEOMETRY_OUTPUT_TYPE_EXT=GL_TRIANGLES
// SetParameter GL_GEOMETRY_VERTICES_OUT_EXT=6

#version 130
#extension GL_EXT_geometry_shader4 : enable
#extension GL_EXT_gpu_shader4 : enable

// ModelMatrix = viewInv * ModelView;
uniform mat4 baseInvModelMatrix;//
uniform mat4 modelInvBaseMatrix;//
uniform int twistMode;

in vec3 positionIn[];
in float pointSizeIn[];
in vec4 colorIn[];
in vec4 colorDiffuseIn[];
in vec4 colorSpecularIn[];
out vec3 position;
out vec4 color;
out vec4 colorDiffuse;
out vec4 colorSpecular;
out vec4 normal;
out vec4 tangent;
out vec4 binormal;
flat in int blendIn[];
flat in int pseudoSphereIn[];
flat in int textureImageIndexIn[];
flat in int twistIDIn[];

flat out int blend;
flat out int pseudoSphere;
flat out int textureImageIndex;

void main(void){

    int length = gl_PositionIn.length();
    
    if (twistMode == 0){
	for (int i = 0; i < length; i++){
	
	    vec4 baseCoord = baseInvModelMatrix * gl_PositionIn[i];
	
	    float x = baseCoord[0];
	    float y = baseCoord[1];
	    float z = baseCoord[2];
	
	    float norm = sqrt(x*x+y*y+z*z);
	    if (norm == 0) continue;

	    x /= norm;
	    y /= norm;
	    z /= norm;
	
	    float sqxy = sqrt(x*x+y*y);
       
	    vec4 normalBase = vec4(x, y, z, 0);
	    vec4 tangentBase = -vec4(-y/sqxy, x/sqxy, 0, 0);
	    vec4 binormalBase = -vec4(-x*z/sqxy, -y*z/sqxy, sqxy, 0);

	    normal = modelInvBaseMatrix * normalBase;
	    tangent  = modelInvBaseMatrix * tangentBase;
	    binormal = modelInvBaseMatrix * binormalBase;

	    vec4 offset[4] = vec4[]((-tangent-binormal), (+tangent-binormal),
				    (-tangent+binormal), (+tangent+binormal));

	    //normal = vec4(gl_NormalMatrix*normal.xyz, 0);
	    //tangent = vec4(gl_NormalMatrix*tangent.xyz, 0);
	    //binormal = vec4(gl_NormalMatrix*binormal.xyz, 0);

	    normal = gl_ModelViewMatrix*normal;
	    tangent = gl_ModelViewMatrix*tangent;
	    binormal = gl_ModelViewMatrix*binormal;
	
	    blend = blendIn[i];
	    pseudoSphere = pseudoSphereIn[i];
	    textureImageIndex = textureImageIndexIn[i];
	    position = positionIn[i];
	    color = colorIn[i];
	    colorDiffuse = colorDiffuseIn[i];
	    colorSpecular = colorSpecularIn[i];

	    float psize = pointSizeIn[i];
	    //psize = 0.1;
	
	    if (color[3] > 0){
		gl_TexCoord[0].st = vec2(1, 0);
		gl_Position = gl_ModelViewProjectionMatrix * (gl_PositionIn[i] + offset[0] * psize) ;
		EmitVertex();
	    
		gl_TexCoord[0].st = vec2(0, 0);
		gl_Position = gl_ModelViewProjectionMatrix * (gl_PositionIn[i] + offset[1] * psize);
		EmitVertex();
	    
		gl_TexCoord[0].st = vec2(1, 1);
		gl_Position = gl_ModelViewProjectionMatrix * (gl_PositionIn[i] + offset[2] * psize);
		EmitVertex();
	    
		gl_TexCoord[0].st = vec2(1, 1);
		gl_Position = gl_ModelViewProjectionMatrix * (gl_PositionIn[i] + offset[2] * psize);
		EmitVertex();
	    
		gl_TexCoord[0].st = vec2(0, 0);
		gl_Position = gl_ModelViewProjectionMatrix * (gl_PositionIn[i] + offset[1] * psize);
		EmitVertex();
	    
		gl_TexCoord[0].st = vec2(0, 1);
		gl_Position = gl_ModelViewProjectionMatrix * (gl_PositionIn[i] + offset[3] * psize);
		EmitVertex();
	    }
	}
	
	EndPrimitive();
    }
    else{
	float rad0 = -3.14159/4;
	float rad1 = 3.14159/4;
	float rad2 = 3.14159*3/4;
	float rad3 = 3.14159*5/4;

	float radius = sqrt(2)/2;
	for (int i = 0; i < length; i++){
	
	    vec4 baseCoord = baseInvModelMatrix * gl_PositionIn[i];
	
	    float x = baseCoord[0];
	    float y = baseCoord[1];
	    float z = baseCoord[2];
	
	    float norm = sqrt(x*x+y*y+z*z);
	    if (norm == 0) continue;

	    x /= norm;
	    y /= norm;
	    z /= norm;
	
	    float sqxy = sqrt(x*x+y*y);
       
	    vec4 normalBase = vec4(x, y, z, 0);
	    vec4 tangentBase = -vec4(-y/sqxy, x/sqxy, 0, 0);
	    vec4 binormalBase = -vec4(-x*z/sqxy, -y*z/sqxy, sqxy, 0);

	    normal = modelInvBaseMatrix * normalBase;
	    tangent  = modelInvBaseMatrix * tangentBase;
	    binormal = modelInvBaseMatrix * binormalBase;

	    vec4 offset[4] = vec4[]((-tangent-binormal), (+tangent-binormal),
				    (-tangent+binormal), (+tangent+binormal));

	    //normal = vec4(gl_NormalMatrix*normal.xyz, 0);
	    //tangent = vec4(gl_NormalMatrix*tangent.xyz, 0);
	    //binormal = vec4(gl_NormalMatrix*binormal.xyz, 0);

	    normal = gl_ModelViewMatrix*normal;
	    tangent = gl_ModelViewMatrix*tangent;
	    binormal = gl_ModelViewMatrix*binormal;
	
	    blend = blendIn[i];
	    pseudoSphere = pseudoSphereIn[i];
	    textureImageIndex = textureImageIndexIn[i];
	    position = positionIn[i];
	    color = colorIn[i];
	    colorDiffuse = colorDiffuseIn[i];
	    colorSpecular = colorSpecularIn[i];
	    float twist = float(twistIDIn[i] % 1000);

	    float psize = pointSizeIn[i];
	    //psize = 0.1;
	
	    if (color[3] > 0){
		gl_TexCoord[0].st = vec2(radius*cos(rad0+twist)+0.5, radius*sin(rad0+twist)+0.5);
		gl_Position = gl_ModelViewProjectionMatrix * (gl_PositionIn[i] + offset[0] * psize) ;
		EmitVertex();
	    
		gl_TexCoord[0] = vec4(radius*cos(rad3+twist)+0.5, radius*sin(rad3+twist)+0.5, 0, 0);
		gl_Position = gl_ModelViewProjectionMatrix * (gl_PositionIn[i] + offset[1] * psize);
		EmitVertex();
	    
		gl_TexCoord[0] = vec4(radius*cos(rad1+twist)+0.5, radius*sin(rad1+twist)+0.5, 0, 0);
		gl_Position = gl_ModelViewProjectionMatrix * (gl_PositionIn[i] + offset[2] * psize);
		EmitVertex();
	    
		gl_TexCoord[0] = vec4(radius*cos(rad1+twist)+0.5, radius*sin(rad1+twist)+0.5, 0, 0);
		gl_Position = gl_ModelViewProjectionMatrix * (gl_PositionIn[i] + offset[2] * psize);
		EmitVertex();
	    
		gl_TexCoord[0] = vec4(radius*cos(rad3+twist)+0.5, radius*sin(rad3+twist)+0.5, 0, 0);
		gl_Position = gl_ModelViewProjectionMatrix * (gl_PositionIn[i] + offset[1] * psize);
		EmitVertex();
	    
		gl_TexCoord[0] = vec4(radius*cos(rad2+twist)+0.5, radius*sin(rad2+twist)+0.5, 0, 0);
		gl_Position = gl_ModelViewProjectionMatrix * (gl_PositionIn[i] + offset[3] * psize);
		EmitVertex();
	    }
	}
	
	EndPrimitive();
    }

    
}
